"""Evaluation module"""
import time
import math
import numpy as np
import cv2
import urllib3
import requests
from bs4 import BeautifulSoup
from stereomideval import image_resize, puttext_multiline
from stereomideval.structures import DatasetType, EvaluationData
from stereomideval.exceptions import ImageSizeNotEqual, InvalidSceneName
from stereomideval.dataset import Dataset


class WebscrapeMiddlebury:
    """Get table data from a webpage"""
    def __init__(self):
        """Initalise WebscapeMiddlebury class"""

    @staticmethod
    def get_web_scene_name_suffix(dataset_type):
        """Get web scene name suffix"""
        scene_name_suffix = ""
        if not dataset_type == DatasetType.imperfect:
            scene_name_suffix = dataset_type
        return scene_name_suffix

    @staticmethod
    def get_table_url(test_or_training, sparsity, metric, mask, has_invalid):
        """
        Get url for table on the Middlebury website using parameters

        Parameters:
            test_or_training (string): table should be for 'test' or 'training' set
            sparsity (string): table should be for 'dense' or 'sparse' sparsity
            metric (Eval.Metric): table should be for given metric
            mask (string): table should be for mask 'nonocc' or 'all'
            has_invalid (bool): table should include invalid results

        Returns:
            url (string): url to table based on given parameters
        """
        # define valid inputs for parameters
        sparse_options = ["dense", "sparse"]
        test_train_options = ["test", "training"]
        metric_options = Metric.get_metrics_list()
        mask_options = ["nonocc", "all"]
        has_invalid_options = ["true", "false"]

        # check validitiy of parameters
        if isinstance(has_invalid, bool):
            if has_invalid:
                has_invalid = "true"
            else:
                has_invalid = "false"
        else:
            if has_invalid not in has_invalid_options:
                raise Exception("has_invalid MUST be {}".format(has_invalid_options))
        if test_or_training not in test_train_options:
            raise Exception("test_or_training MUST be {}".format(test_train_options))
        if sparsity not in sparse_options:
            raise Exception("sparsity MUST be {}".format(sparse_options))
        if metric not in metric_options:
            raise Exception("metric MUST be {}".format(metric_options))
        if mask not in mask_options:
            raise Exception("mask MUST be {}".format(mask_options))

        # define url
        base_url = "https://vision.middlebury.edu/stereo/eval3/table.php"
        base_url += "?dbfile=../results3/results.db"
        db_arg_url = "&type={}&sparsity={}&stat={}&mask={}&hasInvalid={}&ids="
        # fill url with parameters
        url = base_url+db_arg_url.format(test_or_training, sparsity, metric, mask, has_invalid)
        return url

    @staticmethod
    def get_av_metric_vals(metric, dense=True):
        """
        Get list of metrics values from Middlebury website

        Currently assumes using training set with 'all' mask

        Parameters:
            metric (Eval.Metric): metric list to get values for
            scene_name (string): scene name to get values for

        Returns:
            alg_names (list(string)): list of algorithm names
            metric_list (list(string)): list of values for given metric
        """
        dense_or_sparse = "dense"
        if not dense:
            dense_or_sparse = "sparse"
        # Get url for table on Middlebury website
        url = WebscrapeMiddlebury.get_table_url("training", dense_or_sparse, metric, "all", False)
        # disable warnings to avoid message:
        # 'InsecureRequestWarning: Unverified HTTPS request
        # is being made to host 'vision.middlebury.edu'.
        # Adding certificate verification is strongly advised.'
        urllib3.disable_warnings()
        # Load url
        print("Loading results from Middlebury website...")
        page = requests.get(url, verify=False)
        # Parse webpage
        soup = BeautifulSoup(page.content, 'html.parser')
        # Get stereo table from webpage
        table = soup.find('table', {'id': 'stereoTable'})
        # Find table body
        table_body = table.find('tbody')
        # Find table body rows
        table_rows = table_body.findAll('tr')
        # Initalise lists for storing result of table
        alg_names_list = []
        metric_average_list = []
        # (e.g. trying to compare image that is not on website table)
        # Iteration through rows in table
        for table_row in list(table_rows):
            # Find algorithm name in row data
            alg_name = table_row.find('td', {'class': ['algName']})
            # Find overall metric average in row data
            metric_average = table_row.find('td', {'class': ['data wtavg']},
                                            partial=False)
            # Check if metric average was found
            if metric_average is None:
                # Find metric average with 'firstPlace'
                # (the first row in the data has this extra class definition)
                metric_average = table_row.find('td', {"class": ["data wtavg firstPlace"]},
                                                partial=False)
            # Check metric value and algorithm name were found
            if alg_name is not None and metric_average is not None:
                # Add metric value and algorithm name to lists
                metric_average_list.append(float(metric_average.string))
                alg_names_list.append(alg_name.string)
            else:
                raise Exception("Failed to find table data for metric average or algorithm name")
        return alg_names_list, metric_average_list

    @staticmethod
    def get_metric_vals(metric, scene_name, dataset_type=DatasetType.imperfect, dense=True):
        """
        Get list of metrics values from Middlebury website

        Currently assumes using training dense set with nonocc mask

        Parameters:
            metric (Eval.Metric): metric list to get values for
            scene_name (string): scene name to get values for

        Returns:
            alg_names (list(string)): list of algorithm names
            metric_list (list(string)): list of values for given metric
        """
        # Check scene name is valid
        InvalidSceneName.validate_scene_info_list(scene_name, Dataset.get_training_scene_list())
        dense_or_sparse = "dense"
        if not dense:
            dense_or_sparse = "sparse"
        # Get url for table on Middlebury website
        url = WebscrapeMiddlebury.get_table_url("training", dense_or_sparse,
                                                metric, "nonocc", False)
        # disable warnings to avoid message:
        # 'InsecureRequestWarning: Unverified HTTPS request
        # is being made to host 'vision.middlebury.edu'.
        # Adding certificate verification is strongly advised.'
        urllib3.disable_warnings()
        # Load url
        print("Loading results from Middlebury website...")
        page = requests.get(url, verify=False)
        # Parse webpage
        soup = BeautifulSoup(page.content, 'html.parser')
        # Get stereo table from webpage
        table = soup.find('table', {'id': 'stereoTable'})
        # Find table body
        table_body = table.find('tbody')
        # Find table body rows
        table_rows = table_body.findAll('tr')
        # Initalise lists for storing result of table
        alg_names_list = []
        metric_list = []
        web_scene_name_suffix = WebscrapeMiddlebury.get_web_scene_name_suffix(dataset_type)
        # (e.g. trying to compare image that is not on website table)
        # Iteration through rows in table
        for table_row in list(table_rows):
            # Find algorithm name in row data
            alg_name = table_row.find('td', {'class': ['algName']})
            # Find metric value in row data
            metric_val = table_row.find('td', {
                "class": ["{} data datasetCol".format(scene_name+web_scene_name_suffix)]},
                partial=False)
            # Check if metric value was found
            if metric_val is None:
                # Find metric value with 'firstPlace'
                # (the first row in the data has this extra class definition)
                metric_val = table_row.find('td', {
                    "class": ["{} firstPlace data datasetCol".format(
                        scene_name+web_scene_name_suffix)]},
                    partial=False)
            # Check metric value and algorithm name were found
            if metric_val is not None and alg_name is not None:
                # Add metric value and algorithm name to lists
                metric_list.append(float(metric_val.string))
                alg_names_list.append(alg_name.string)
            else:
                raise Exception("Failed to find table data for metric value or algorithm name")
        return alg_names_list, metric_list


class Timer:
    """
    Timer

    Simplify getting elapsed time for tasks
    """
    def __init__(self):
        self.start_time = None

    def start(self):
        """Start timer"""
        self.start_time = time.time()

    def elapsed(self):
        """
        Get elapsed time

        Returns:
            elapsed_time (float): time since start of time (seconds)
        """
        # Check timer has been started
        if self.start_time is None:
            raise Exception("Timer not running")
        # Calculate elapsed time
        elapsed_time = time.time() - self.start_time
        return elapsed_time


class Metric:
    """
    Metrics used for evaluating data in the Middlebury dataset
    """
    # Define metrics (must match stat value in Middlebury website url)
    bad050 = "bad050"
    bad100 = "bad100"
    bad200 = "bad200"
    bad400 = "bad400"
    avgerr = "avgerr"
    rms = "rms"
    a50 = "A50"
    a90 = "A90"
    a95 = "A95"
    a99 = "A99"
    time = "time"
    time_mp = "time/MP"
    time_gd = "time/Gdisp"
    TIME_METRICS = [time, time_mp, time_gd]
    BAD_METRICS = [bad050, bad100, bad200, bad400]
    QUANTILE_METRICS = [a50, a90, a95, a99]

    @staticmethod
    def get_metrics_list():
        """Get metric strings as a list"""
        m = Metric
        return [
            m.bad050, m.bad100, m.bad200, m.bad400,
            m.avgerr, m.rms,
            m.a50, m.a90, m.a95, m.a99,
            m.time, m.time_mp, m.time_gd
        ]

    @staticmethod
    def get_quantile_percentage(metric):
        """Get percentage value for quantile error metrics"""
        if metric not in Metric.QUANTILE_METRICS:
            raise Exception("Must provided quantile matric e.g. Metric.a50")
        if metric == Metric.a50:
            percentage = 0.5
        elif metric == Metric.a90:
            percentage = 0.9
        elif metric == Metric.a95:
            percentage = 0.95
        elif metric == Metric.a99:
            percentage = 0.99
        return percentage

    @staticmethod
    def get_bad_percentage(metric):
        """Get percentage value for bad pixel error metrics"""
        if metric not in Metric.BAD_METRICS:
            raise Exception("Must provided bad matric e.g. Metric.bad050")
        if metric == Metric.bad050:
            percentage = 0.5
        elif metric == Metric.bad100:
            percentage = 1
        elif metric == Metric.bad200:
            percentage = 2
        elif metric == Metric.bad400:
            percentage = 4
        return percentage

    @staticmethod
    def calc_time_mp(test_data, match_time):
        """
        Time normalised by number of megapixels

        Parameters:
            test_data (numpy): test image (used to get number of megapixels)
            match_time (float): elapsed time for match (in seconds)

        Returns:
            norm_time (float): time normalised by megapixels
        """
        # Count megapixels in image
        num_of_megapixels = test_data.size/1000000
        # Calculate time normalised by megapixels
        norm_time = match_time/num_of_megapixels
        return norm_time

    @staticmethod
    def calc_time_gd(test_data, match_time, ndisp):
        """
        Time normalised by number of disparity hypotheses

        Parameters:
            test_data (numpy): test image (used to get number of megapixels)
            match_time (float): elapsed time for match (in seconds)
            ndisp (int): number of disparities

        Returns:
            norm_time (float): time normalised by number of disparity hypotheses
        """
        # Count gigapixels in image
        num_of_gigapixels = test_data.size/1000000000
        # Calculate time normalised by megapixels
        norm_time = match_time/(num_of_gigapixels*ndisp)
        return norm_time

    @staticmethod
    def calc_quantile(ground_truth, test_data, quantile):
        """
        Q-th quantile of the data error

        Parameters:
            ground_truth (numpy): 2D ground truth image to use for comparision
            test_data (numpy): 2D test image to compare against ground truth
            quantile (float): q-th quantile of the data

        Returns:
            quantile_error (float): Q-th quantile of the data error
        """
        diff = Metric.calc_diff(ground_truth, test_data)
        # Get the absolute difference (positive only)
        abs_diff = np.abs(diff)
        quantile_error = np.quantile(abs_diff, quantile)
        return quantile_error

    @staticmethod
    def calc_avgerr(ground_truth, test_data):
        """
        Average error

        Parameters:
            ground_truth (numpy): 2D ground truth image to use for comparision
            test_data (numpy): 2D test image to compare against ground truth
        Returns:
            average (float): average error in test data
        """
        diff = Metric.calc_diff(ground_truth, test_data)
        # Get the absolute difference (positive only)
        abs_diff = np.abs(diff)
        average = np.average(abs_diff)
        return average

    @staticmethod
    def calc_diff(ground_truth, test_data):
        """
        Difference between ground truth and test data

        Parameters:
            ground_truth (numpy): 2D ground truth image to use for comparision
            test_data (numpy): 2D test image to compare against ground truth
        Returns:
            diff (numpy): test data subtracted from ground truth
        """
        # Check images are the same size
        ImageSizeNotEqual.validate(test_data, ground_truth)
        # Replace nan and inf values with zero
        test_data_no_nan = np.nan_to_num(test_data, nan=0.0, posinf=0.0, neginf=0.0)
        ground_truth_no_nan = np.nan_to_num(ground_truth, nan=0.0, posinf=0.0, neginf=0.0)
        # Subtract test data from ground truth to find difference
        diff = np.subtract(test_data_no_nan, ground_truth_no_nan)
        return diff

    @staticmethod
    def calc_bad_pix_error(ground_truth, test_data, threshold=2.0):
        """
        Bad pixel error

        Calcuate percentage of bad pixels between test data and ground truth
        A bad pixel is defined as a disparity with an error larger than the threshold

        Parameters:
            ground_truth (numpy): 2D ground truth image to use for comparision
            test_data (numpy): 2D test image to compare against ground truth
            threshold (float): Threshold below which to classify a bad pixel

        Returns:
            perc_bad (float): Bad pixel percentage error in test data.
        """
        ImageSizeNotEqual.validate(test_data, ground_truth)
        # Calculate pixel difference between ground truth and test data
        diff = Metric.calc_diff(ground_truth, test_data)
        # Get the absolute difference (positive only)
        abs_diff = np.abs(diff)
        # Count number of 'bad' pixels
        bad_count = float((~(abs_diff < threshold)).sum())
        # Convert number of 'bad' pixels to percentage
        total_size = float(ground_truth.shape[0] * ground_truth.shape[1])
        perc_bad = (bad_count/total_size)*100
        return perc_bad

    @staticmethod
    def calc_rmse(ground_truth, test_data):
        """
        Root mean squared error

        Calculate root mean squared error between test data and ground truth

        Parameters:
            ground_truth (numpy): 2D ground truth image to use for comparision
            test_data (numpy): 2D test image to compare against ground truth

        Returns:
            err (float): Root mean squared error of two images,
                the lower the error, the more "similar" the two images
        """
        # Check test data and ground truth are the same size
        ImageSizeNotEqual.validate(test_data, ground_truth)
        # Calculate the root of mse
        rmse = math.sqrt(Metric.calc_mse(ground_truth, test_data))
        return rmse

    @staticmethod
    def calc_mse(ground_truth, test_data):
        """
        Mean squared error

        Calculate the mean squared error between test data and ground truth

        Parameters:
            ground_truth (numpy): 2D ground truth image to use for comparision
            test_data (numpy): 2D test image to compare against ground truth

        Returns:
            err (float): Mean squared error of two images,
                the lower the error, the more "similar" the two images
        """
        ImageSizeNotEqual.validate(test_data, ground_truth)
        # Calculate difference (error) between ground truth and test data
        diff = Metric.calc_diff(ground_truth, test_data)
        # Calculate mean of the square error
        err = np.square(diff).mean()
        return err


class Eval:
    """Evaluate disparity image against ground truth"""
    @staticmethod
    def display_results(match_result, eval_result_list=None, window_name="Results", wait=1000):
        """Display match results and evaluation to OpenCV window"""
        match_result_image = match_result.match_result
        ground_truth = match_result.ground_truth
        left_image = match_result.left_image
        right_image = match_result.right_image
        # remove negative disparities
        match_result_image[match_result_image < 0] = 0.0
        ground_truth[ground_truth < 0] = 0.0
        # Replace nan and inf values with zero
        match_result_image = np.nan_to_num(match_result_image, nan=0.0, posinf=0.0, neginf=0.0)
        ground_truth = np.nan_to_num(ground_truth, nan=0.0, posinf=0.0, neginf=0.0)
        # normalise image
        match_result_image = cv2.normalize(match_result_image, None, 0, 255, cv2.NORM_MINMAX)
        ground_truth = cv2.normalize(ground_truth, None, 0, 255, cv2.NORM_MINMAX)
        # convert to uint8 (required by applyColorMap function)
        match_result_image = match_result_image.astype(np.uint8)
        ground_truth = ground_truth.astype(np.uint8)
        # apply colormap
        match_result_image = cv2.applyColorMap(match_result_image, cv2.COLORMAP_JET)
        ground_truth = cv2.applyColorMap(ground_truth, cv2.COLORMAP_JET)
        # Resize colormap image for displaying in window
        match_result_image = image_resize(match_result_image, width=320)
        ground_truth = image_resize(ground_truth, width=320)
        left_image = image_resize(left_image, width=320)
        right_image = image_resize(right_image, width=320)
        # Concatinate images
        disp_images = cv2.hconcat([ground_truth, match_result_image])
        stereo_image = cv2.hconcat([left_image, right_image])
        display_image = cv2.vconcat([stereo_image, disp_images])

        if eval_result_list is not None:
            # Print metrics on display image
            msg = ""
            for eval_result in eval_result_list:
                metric_result = eval_result.result
                metric_rank = eval_result.rank
                metric = eval_result.metric
                if metric_result is not None:
                    msg += metric+" {:.2f} ".format(metric_result)
                    if metric_rank is not None:
                        msg += "({})\n".format(metric_rank)
                    else:
                        msg += "\n"
            display_image = puttext_multiline(
                img=display_image, text=msg, org=(10, 10), font=cv2.FONT_HERSHEY_TRIPLEX,
                font_scale=0.7, color=(0, 0, 0), thickness=1, outline_color=(255, 255, 255))

        # Display in OpenCV window
        cv2.imshow(window_name, display_image)
        cv2.waitKey(wait)

    @staticmethod
    def evaluate_match_result_list(match_result_list, display_results=False,
                                   display_window_name="Results", display_wait=1000):
        """
        Evaluate match result list

        Paramaters:
            match_result_list (list(MatchResult)): list of match results
            display_results (bool): display results to OpenCV window while evaluating
            display_window (string): name of OpenCV window
            display_wait (int): time to wait, used in 'cv2.waitKey(display_wait)'

        Returns:
            metric_average_list (list(MatchResult)): list of average metrics across all test data
            eval_result_list_list (list(list(MatchResult))): list of lists of match results,
                                                        seperate list for each test data provided
        """
        eval_result_list_list = []
        for match_result in match_result_list:
            # Evaluate test data against all metrics
            eval_result_list = Eval.eval_all_metrics(match_result)
            if display_results:
                # Display evaluation results to OpenCV window
                Eval.display_results(match_result, eval_result_list,
                                     display_window_name, display_wait)
            eval_result_list_list.append(eval_result_list)

        metric_average_list = \
            Eval.average_all_metrics_across_scenes(eval_result_list_list)

        print("Metric average over all test data:")
        for metric_average in metric_average_list:
            msg = "{}: {} ({})"
            msg = msg.format(metric_average.metric, metric_average.result, metric_average.rank)
            print(msg)

        return metric_average_list, eval_result_list_list

    @staticmethod
    def evaluate_match_data_list(match_data_list, get_metric_rank=False, get_av_metric_rank=False,
                                 dense=True, display_results=False, display_window_name="Results",
                                 display_wait=1000):
        """
        Evaluate match result list

        Paramaters:
            match_result_list (list(MatchData)): list of match data
            get_metric_rank (bool): compare evaluation data against Middlesbury website ranking
            get_av_metric_rank (bool): compare average evaluation data against Middlesbury
                                         website ranking
            display_results (bool): display results to OpenCV window while evaluating
            display_window (string): name of OpenCV window
            display_wait (int): time to wait, used in 'cv2.waitKey(display_wait)'

        Returns:
            metric_average_list (list(MatchResult)): list of average metrics across all test data
            eval_data_list (list(EvaluationData)): list of evaluation data, seperate list for
                                                    each test data provided
        """
        eval_data_list = []
        eval_result_list_list = []
        for match_data in match_data_list:
            scene_info = match_data.scene_info

            if get_metric_rank:
                # Evaluate test data against all metrics
                eval_result_list = Eval.eval_all_metrics_rank(match_data, dense)
            else:
                # Evaluate test data against all metrics
                eval_result_list = Eval.eval_all_metrics(match_data.match_result)

            if display_results:
                Eval.display_results(match_data.match_result, eval_result_list,
                                     display_window_name, display_wait)
            eval_data = EvaluationData(scene_info, eval_result_list)
            eval_data_list.append(eval_data)
            eval_result_list_list.append(eval_result_list)

        if get_av_metric_rank:
            metric_average_list = \
                Eval.average_all_metrics_across_scenes_rank(eval_data_list, dense)
        else:
            metric_average_list = \
                Eval.average_all_metrics_across_scenes(eval_result_list_list)

        print("Weighted metric average over all test data:")
        for metric_average in metric_average_list:
            msg = "{}: {} ({})"
            msg = msg.format(metric_average.metric, metric_average.result, metric_average.rank)
            print(msg)

        return metric_average_list, eval_data_list

    @staticmethod
    def get_average_metric_rank(value, metric, dense=True):
        """
        Compare new value to middlebury metric and
        calculate it's rank when compared to the current
        table of results from the Middlebury website

        Parameters:
            metric (Eval.Metrics): metric list to search against
            scene_name (string): name of scene from Middlebury stereo dataset (2014)
            value (float): new value to compare against online metric list

        Returns:
            rank (int): rank position of value when compared to metric list
        """
        _, metric_avs = WebscrapeMiddlebury.get_av_metric_vals(metric, dense)
        if len(metric_avs) <= 0:
            raise Exception("Failed to receive metric averages from Middlebury webpage")
        print("Comparing against {} results...".format(len(metric_avs)))
        # add your metric value to the list
        metric_avs.append(value)
        metric_ranks = Eval.rank_vals(metric_avs)
        # get last element in list (this will be the element we added)
        return metric_ranks[-1]

    @staticmethod
    def get_metric_rank(metric, scene_name, value, dataset_type=DatasetType.imperfect, dense=True):
        """
        Compare new value to middlebury metric and
        calculate it's rank when compared to the current
        table of results from the Middlebury website

        Parameters:
            metric (Eval.Metrics): metric list to search against
            scene_name (string): name of scene from Middlebury stereo dataset (2014)
            value (float): new value to compare against online metric list

        Returns:
            rank (int): rank position of value when compared to metric list
        """
        _, metric_vals = WebscrapeMiddlebury.get_metric_vals(metric, scene_name,
                                                             dataset_type, dense)
        if len(metric_vals) <= 0:
            raise Exception("Failed to receive metric values from Middlebury webpage")
        print("Comparing against {} results...".format(len(metric_vals)))
        # add your metric value to the list
        metric_vals.append(value)
        metric_ranks = Eval.rank_vals(metric_vals)
        # get last element in list (this will be the element we added)
        return metric_ranks[-1]

    @staticmethod
    def rank_vals(data):
        """
        Rank values in list from 1 to n based on value (1 is lowest)

        Parameters:
            data (list(float)): list of values to rank

        Returns:
            rank (list(int)): list of ranks for values
        """
        # Get data length
        data_len = len(data)
        ivec = sorted(range(len(data)), key=data.__getitem__)
        svec = [data[rank] for rank in ivec]
        sumranks = 0
        dupcount = 0
        newarray = [0]*data_len
        for data_index in range(data_len):
            sumranks += data_index
            dupcount += 1
            if data_index == data_len-1 or svec[data_index] != svec[data_index+1]:
                averank = sumranks / float(dupcount) + 1
                for j in range(data_index-dupcount+1, data_index+1):
                    newarray[ivec[j]] = int(averank)
                sumranks = 0
                dupcount = 0
        return newarray

    @staticmethod
    def average_all_metrics_across_scenes_rank(evaluation_data_list, dense=True):
        """
        Evaluate all metrics across all scenes and compare rank

        Parameters:
            evaluation_data_list (list(EvaluationData)): list of evaluation data
            dense (bool): compare to dense online ranking table

        Returns:
            metric_average_list (list(EvaluationResult)): average result and rank across all scenes
        """
        # Initalise dictionary
        metric_average_list = []
        # Run evaluation for each metric
        for metric in Metric.get_metrics_list():
            metric_average = \
                Eval.average_metric_across_scenes_rank(metric, evaluation_data_list, dense)
            metric_average_list.append(metric_average)

        return metric_average_list

    @staticmethod
    def average_all_metrics_across_scenes(evaluation_result_list_list):
        """
        Evaluate all metrics across all scenes

        Parameters:
            evaluation_data_list (list(list(EvaluationResults))):
                list of lists of evaluation results
        Returns:
            metric_average_list (list(EvaluationResult)): average result across all scenes
        """
        # Initalise dictionary
        metric_average_list = []
        # Run evaluation for each metric
        for metric in Metric.get_metrics_list():
            metric_average = Eval.average_metric_across_scenes(metric, evaluation_result_list_list)
            metric_average_list.append(metric_average)
        return metric_average_list

    @staticmethod
    def average_metric_across_scenes_rank(metric, evaluation_data_list, dense=True):
        """
        Evaluate metric result and rank across all scenes

        Parameters:
            metric (Metric): metric to evaluate
            evaluation_data_list (list(EvaluationData)):
                list of evaluation data
            dense (bool): compare to dense online ranking table
        Returns:
            average_metric_result (list(EvaluationResult)):
                average result for metric across all scenes
        """
        metric_result_list = []
        for evaluation_data in evaluation_data_list:
            scene_info = evaluation_data.scene_info
            evaluation_result_list = evaluation_data.eval_result_list
            metric_found = False
            for evaluation_result in evaluation_result_list:
                if evaluation_result.metric == metric:
                    metric_result_list.append(evaluation_result)
                    metric_found = True
            # raise exception if missing metric
            if not metric_found:
                raise Exception("Missing metric {} in evaluation data".format(metric))

        average = 0
        scene_count = 0
        for metric_result in metric_result_list:
            metric_result_val = metric_result.result
            if metric_result_val is None:
                raise Exception("Missing metric result for {}".format(metric))
            average += metric_result_val * scene_info.weight
            scene_count += 1
        average = average / scene_count
        average_metric_result = EvaluationData.EvaluationResult(metric, average)

        if average_metric_result.result is None:
            raise Exception("Failed to find result value for {} metric".format(metric))

        # Compare result with Middlebury results and return rank
        msg = "Comparing average {} result of {:.2f} with online results..."
        msg = msg.format(metric, average_metric_result.result)
        print(msg)
        average_metric_result.rank = \
            Eval.get_average_metric_rank(average_metric_result.result, metric, dense)
        print("Average {} result of {:.2f} is rank {}".format(
            metric, average_metric_result.result, average_metric_result.rank))
        return average_metric_result

    @staticmethod
    def average_metric_across_scenes(metric, evaluation_result_list_list):
        """
        Evaluate metric result across all scenes

        Parameters:
            metric (Metric): metric to evaluate
            evaluation_data_list (list(list(EvaluationResults))):
                list of lists of evaluation results
        Returns:
            average_metric_result (list(EvaluationResult)):
                average result for metric across all scenes
        """
        metric_result_list = []
        for evaluation_result_list in evaluation_result_list_list:
            metric_found = False
            for evaluation_result in evaluation_result_list:
                if evaluation_result.metric == metric:
                    metric_result_list.append(evaluation_result)
                    metric_found = True
            # raise exception if missing metric
            if not metric_found:
                raise Exception("Missing metric {} in evaluation data".format(metric))

        average = 0
        scene_count = 0
        evaluation_result_list = []
        for metric_result in metric_result_list:
            metric_result_val = metric_result.result
            if metric_result_val is None:
                raise Exception("Missing metric result for {}".format(metric))
            average += metric_result_val
            scene_count += 1
        average = average / scene_count
        average_metric_result = EvaluationData.EvaluationResult(metric, average)
        return average_metric_result

    @staticmethod
    def eval_all_metrics_rank(match_data, dense=True):
        """
        Evaluate test data using all Middlesbury metrics and compare rank with online results

        Parameters:
            ground_truth (numpy): 2D ground truth image
            test_data (numpy): 2D test image
            elapsed_time (float): time to complete match (only used in time metrics)
            TODO

        Returns:
            TODO
        """
        # Initalise dictionary
        metric_result_list = []
        # Run evaluation for each metric
        for metric in Metric.get_metrics_list():
            # Evaluate metric and compare rank
            metric_result = Eval.eval_metric_rank(metric, match_data, dense)
            # Add metric result to list
            metric_result_list.append(metric_result)
        return metric_result_list

    @staticmethod
    def eval_all_metrics(match_result):
        """
        Evaluate test data using all Middlesbury metrics

        Parameters:
            ground_truth (numpy): 2D ground truth image
            test_data (numpy): 2D test image
            elapsed_time (float): time to complete match (only used in time metrics)

        Returns:
            TODO
        """
        # Initalise dictionary
        metric_result_list = []
        # Run evaluation for each metric
        for metric in Metric.get_metrics_list():
            # Evaluate metric and compare rank
            metric_result = Eval.eval_metric(metric, match_result)
            # Add metric result to list
            metric_result_list.append(metric_result)
        return metric_result_list

    @staticmethod
    def eval_metric_rank(metric, match_data, dense=True):
        """
        Run evaluation routine based on metric and compare rank with online results

        Parameters:
            metric (Eval.Metrics): Evaluation metrics to calculate
            match_data (Structures.MatchData): match data (see Structures.py for details)
            dense (bool): compare to dense online ranking table

        Returns:
            metric_result (Structures.EvaluationResult): metric result
        """
        # Check metric is valid
        if metric not in Metric.get_metrics_list():
            raise Exception("Invalid metric")
        scene_info = match_data.scene_info
        # Run evaluation routine based on metric chosen
        metric_result = Eval.eval_metric(metric, match_data.match_result)

        if metric_result.result is None:
            raise Exception("Failed to evaluate metric {}".format(metric))

        # Compare result with Middlebury results and return rank
        print("Comparing {} result of {:.2f} with online results...".format(
            metric, metric_result.result))
        scene_info = match_data.scene_info
        metric_result.rank = Eval.get_metric_rank(
            metric, scene_info.scene_name, metric_result.result,
            scene_info.dataset_type, dense)

        if metric_result.rank is None:
            raise Exception("Failed to get {} rank".format(metric))

        print("{} result of {:.2f} is rank {}".format(
            metric, metric_result.result, metric_result.rank))

        return metric_result

    @staticmethod
    def eval_metric(metric, match_result):
        """
        Run evaluation routine based on metric

        Parameters:
            metric (Eval.Metrics): Evaluation metrics to calculate
            match_data (Structures.MatchData): match data (see Structures.py for details)

        Returns:
            metric_result (Structures.EvaluationResult): metric result
        """
        # TODO fix 'PEP8 C901 function is too complex (12)'
        # Check metric is valid
        if metric not in Metric.get_metrics_list():
            raise Exception("Invalid metric")
        # Check match time parameter is defined when using time metrics
        if match_result.match_time is None:
            raise Exception("Match time parameter is missing")
        # Initalise return variable
        result = None
        print("Evaluating {} metric...".format(metric))
        # Run evaluation routine based on metric chosen
        if metric in Metric.BAD_METRICS:
            bad_perc = Metric.get_bad_percentage(metric)
            result = Metric.calc_bad_pix_error(
                match_result.ground_truth, match_result.match_result, bad_perc)
        elif metric in Metric.QUANTILE_METRICS:
            quantile_perc = Metric.get_quantile_percentage(metric)
            result = Metric.calc_quantile(
                match_result.ground_truth, match_result.match_result, quantile_perc)
        elif metric in Metric.TIME_METRICS:
            if metric == Metric.time:
                result = match_result.match_time
            elif metric == Metric.time_mp:
                result = Metric.calc_time_mp(match_result.match_result, match_result.match_time)
            elif metric == Metric.time_gd:
                if match_result.ndisp is None:
                    raise Exception("number of disparities parameter is missing")
                result = Metric.calc_time_gd(match_result.match_result,
                                             match_result.match_time,
                                             match_result.ndisp)
        elif metric == Metric.rms:
            result = Metric.calc_rmse(match_result.ground_truth, match_result.match_result)
        elif metric == Metric.avgerr:
            result = Metric.calc_avgerr(match_result.ground_truth, match_result.match_result)

        metric_result = EvaluationData.EvaluationResult(metric, result)
        return metric_result



"""Dataset module"""
import os
import sys
import zipfile
import re
import glob
import numpy as np
import cv2
import wget
from stereomideval.structures import DatasetType, CalibrationData, TestData, SceneInfo
from stereomideval.exceptions import PathNotFound, MalformedPFM, InvalidSceneName


class Dataset:
    """Download and read data from stereo Middlesbury dataset (2014)"""

    # List of available scenes in the Middlebury stereo dataset (2014)
    Adirondack = "Adirondack"
    Backpack = "Backpack"
    Bicycle1 = "Bicycle1"
    Books = "Books"
    Cable = "Cable"
    Classroom1 = "Classroom1"
    Couch = "Couch"
    Computer = "Computer"
    Cones = "Cones"
    Dolls = "Dolls"
    Drumsticks = "Drumsticks"
    Dwarves = "Dwarves"
    Flowers = "Flowers"
    Jadeplant = "Jadeplant"
    Laundry = "Laundry"
    Mask = "Mask"
    Moebius = "Moebius"
    Motorcycle = "Motorcycle"
    Piano = "Piano"
    Pipes = "Pipes"
    Playroom = "Playroom"
    Playtable = "Playtable"
    Reindeer = "Reindeer"
    Recycle = "Recycle"
    Shelves = "Shelves"
    Shopvac = "Shopvac"
    Sticks = "Sticks"
    Storage = "Storage"
    Sword1 = "Sword1"
    Sword2 = "Sword2"
    Umbrella = "Umbrella"
    Vintage = "Vintage"
    Art = "Art"
    Teddy = "Teddy"
    STEREO_MIDDLEBURY_SCENES_2014 = [
        Adirondack, Backpack, Bicycle1, Cable, Classroom1,
        Couch, Flowers, Jadeplant, Mask, Motorcycle,
        Piano, Pipes, Playroom, Playtable, Recycle,
        Shelves, Shopvac, Sticks, Storage,
        Sword1, Sword2, Umbrella, Vintage
    ]
    STEREO_MIDDLEBURY_SCENES_2003 = [
        Cones, Teddy
    ]
    STEREO_MIDDLEBURY_SCENES_2005 = [
        Art, Books, Computer, Dolls, Drumsticks,
        Dwarves, Laundry, Moebius, Reindeer
    ]

    STEREO_MIDDLEBURY_SCENES = STEREO_MIDDLEBURY_SCENES_2003 + \
        STEREO_MIDDLEBURY_SCENES_2005 + STEREO_MIDDLEBURY_SCENES_2014

    STEREO_MIDDLEBURY_TRAINING_SCENES = [
        SceneInfo(Adirondack, DatasetType.imperfect, 1.0),
        SceneInfo(Art, DatasetType.lighting_changed, 1.0),
        SceneInfo(Jadeplant, DatasetType.imperfect, 1.0),
        SceneInfo(Motorcycle, DatasetType.imperfect, 1.0),
        SceneInfo(Motorcycle, DatasetType.exposure_changed, 1.0),
        SceneInfo(Piano, DatasetType.imperfect, 1.0),
        SceneInfo(Piano, DatasetType.lighting_changed, 0.5),
        SceneInfo(Pipes, DatasetType.imperfect, 1.0),
        SceneInfo(Playroom, DatasetType.imperfect, 0.5),
        SceneInfo(Playtable, DatasetType.imperfect, 0.5),
        SceneInfo(Playtable, DatasetType.perfect, 1.0),
        SceneInfo(Recycle, DatasetType.imperfect, 1.0),
        SceneInfo(Shelves, DatasetType.imperfect, 0.5),
        SceneInfo(Teddy, DatasetType.imperfect, 1.0),
        SceneInfo(Vintage, DatasetType.imperfect, 0.5),
    ]

    # TODO: add 2014 test scenes (not in organised single zips like test data so requires more work)

    @staticmethod
    def normalise_pfm_data(data, max_val_pct=0.1):
        """
        Normalise disparity pfm image

        Parameters:
            data (numpy): 2D pfm image data to normalise
            max_val_pct (float): TODO what is this?

        Returns:
            norm_pfm_data (numpy): 2D normalised pfm image data
        """
        norm_pfm_data = np.where(data == np.inf, -1, data)
        max_val = np.max(norm_pfm_data)
        max_val += max_val * max_val_pct
        norm_pfm_data = np.where(norm_pfm_data == -1, max_val, norm_pfm_data)
        norm_pfm_data = cv2.normalize(
            norm_pfm_data, dst=None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX)
        norm_pfm_data = norm_pfm_data.astype(np.uint8)
        return norm_pfm_data

    @staticmethod
    def load_cal(cal_filepath):
        """
        Load camera parameters from calibration file

        Parameters:
            cal_filepath (string): filepath to calibration file (usually calib.txt)
                Expected format:
                    cam0=[3997.684 0 1176.728; 0 3997.684 1011.728; 0 0 1]
                    cam1=[3997.684 0 1307.839; 0 3997.684 1011.728; 0 0 1]
                    doffs=131.111
                    baseline=193.001
                    width=2964
                    height=1988
                    ndisp=280
                    isint=0
                    vmin=31
                    vmax=257
                    dyavg=0.918
                    dymax=1.516

        Returns:
            depth (numpy): 2D depth image (units meters)
        """

        # Check calibration file exists
        if not os.path.exists(cal_filepath):
            print("Calibration file not found")
            print(cal_filepath)
            raise PathNotFound(cal_filepath, "Calibration file not found")

        # Open calibration file
        file = open(cal_filepath, 'rb')
        # Read first line
        # expected format: "cam0=[3997.684 0 1176.728; 0 3997.684 1011.728; 0 0 1]"
        cam0_line = file.readline().decode('utf-8').rstrip()
        # Read second line but ignore the data as cam0 and cam1 have the same parameters
        _ = file.readline().decode('utf-8').rstrip()
        # Read third line (expected format: "doffs=131.111")
        doffs_line = file.readline().decode('utf-8').rstrip()
        # Read third line (expected format: "baseline=193.001")
        baseline_line = file.readline().decode('utf-8').rstrip()
        # Read 4th line (expected format: "width=2964")
        width_line = file.readline().decode('utf-8').rstrip()
        # Read 5th line (expected format: "height=19881")
        height_line = file.readline().decode('utf-8').rstrip()
        # Read 6th line (expected format: "ndisp=280")
        ndisp_line = file.readline().decode('utf-8').rstrip()

        # Read all numbers from cam0 line using regex
        nums = re.findall("\\d+\\.\\d+", cam0_line)
        # Get camera parmeters from file data
        cam0_f = float(nums[0])
        cam0_cx = float(nums[1])
        cam0_cy = float(nums[3])

        # Get doffs and baseline from file data
        doffs = float(re.findall("\\d+\\.\\d+", doffs_line)[0])
        baseline = float(re.findall("\\d+", baseline_line)[0])
        width = float(re.findall("\\d+", width_line)[0])
        height = float(re.findall("\\d+", height_line)[0])
        ndisp = float(re.findall("\\d+", ndisp_line)[0])

        cal_data = CalibrationData(width, height, cam0_cx, cam0_cy, cam0_f, doffs, baseline, ndisp)
        return cal_data

    @staticmethod
    def disp_to_depth(disp, focal_length, doffs, baseline):
        """
        Convert from disparity to depth using calibration file

        Parameters:
            disp (numpy): 2D disparity image (result of stereo matching)
            focal_length (float): Focal length of left camera (in pixels)
            doffs (float): x-difference of principal points, doffs = cx1 - cx0
            baseline (float): Baseline distance between left and right camera (in mm)

        Returns:
            depth (numpy): 2D depth image (units meters)
        """
        # Calculate depth from disparitiy
        # Z = baseline * f / (disp + doeff)
        z_mm = baseline * focal_length / (disp + doffs)
        # Z is in mm, convert to meters
        depth = z_mm / 1000
        return depth

    @staticmethod
    def load_pfm(filepath):
        """
        Load pfm data from file

        Parameters:
            filepath (string): filepath to pfm image (e.g. image.pfm)

        Returns:
            pfm_data (numpy): 2D image filled with data from pfm file
        """
        # Check file exists
        if not os.path.exists(filepath):
            raise PathNotFound(filepath, "Pfm file does not exist")
        # Open pfm file
        file = open(filepath, 'rb')

        color = None
        width = None
        height = None
        scale = None
        endian = None

        # Read header to check pf type
        header = file.readline().decode('utf-8').rstrip()
        if header == 'PF':
            color = True
        elif header == 'Pf':
            color = False
        else:
            raise MalformedPFM('Not a PFM file.')

        # Read dimensions from pfm file and check they match expected
        dim_match = re.match(r'^(\d+)\s(\d+)\s$', file.readline().decode('utf-8'))
        if dim_match:
            width, height = map(int, dim_match.groups())
        else:
            raise MalformedPFM('Malformed PFM header')

        # Read data scale from file
        scale = float(file.readline().decode('utf-8').rstrip())
        if scale < 0:  # little-endian
            endian = '<'
            scale = -scale
        else:
            endian = '>'  # big-endian

        # Read image data from pfm file
        data = np.fromfile(file, endian + 'f')
        # Format image data into expected numpy image
        shape = (height, width, 3) if color else (height, width)
        data = np.reshape(data, shape)
        # Flip image vertically as image appears upside-down
        data = cv2.flip(data, 0)
        return data, scale

    @staticmethod
    def load_scene_disparity(scene_name, dataset_folder, display_images=False,
                             display_time=500, load_perfect=False):
        """
        Load disparity image from scene folder

        Parameters:
            scene_name (string): Scene name to load data (use 'get_scene_list()'
                to see possible scenes) Must have already been downloaded to the
                dataset foldering using 'download_scene_data()'
            dataset_folder (string): Path to folder where dataset has been downloaded
            display_image (bool): Optional. Should the scene data be displayed when it is loaded.
            display_time (int): Optional. Length of time to display each image once loaded.

        Returns:
            disp_image (numpy): 2D disparity image
                loaded from scene data (result of stereo matching)
        """
        scene_year = Dataset.get_scene_year(scene_name)
        if scene_year == "2014":
            perfect_suffix = Dataset.get_perfect_suffix(load_perfect)
            # Get name of disparity image (pfm) in folder
            disp_filename = os.path.join(dataset_folder, scene_name+perfect_suffix, "disp0.pfm")
        elif scene_year == "2003" or scene_year == "2005":
            disp_filename = os.path.join(dataset_folder, scene_name, "disp0.png")
        # Check disparity file exists
        if not os.path.exists(disp_filename):
            print("Disparity pfm file does not exist")
            print(disp_filename)
            raise PathNotFound(disp_filename, "Disparity pfm file does not exist")
        # Load disparity file to numpy image
        if scene_year == "2014":
            disp_image, _ = Dataset.load_pfm(disp_filename)
        elif scene_year == "2003" or scene_year == "2005":
            disp_image = cv2.imread(disp_filename, cv2.IMREAD_UNCHANGED)
            disp_image /= 4
        if display_images:
            # Display disparity image in opencv window
            norm_disp_image = Dataset.normalise_pfm_data(disp_image)
            norm_disp_image_resize = cv2.resize(norm_disp_image, dsize=(0, 0), fx=0.2, fy=0.2)
            cv2.imshow('image', cv2.applyColorMap(norm_disp_image_resize, cv2.COLORMAP_JET))
            cv2.waitKey(display_time)
        return disp_image

    @staticmethod
    def get_perfect_suffix(dataset_type):
        """Get perfect suffix"""
        perfect_suffix = "-imperfect"
        if dataset_type == DatasetType.perfect:
            perfect_suffix = "-perfect"
        return perfect_suffix

    @staticmethod
    def get_image_suffix(dataset_type):
        """Get image suffix"""
        image_suffix = ""
        if dataset_type in DatasetType.SCENE_CHANGED_TYPES:
            image_suffix = dataset_type
        return image_suffix

    @staticmethod
    def load_scene_stereo_pair(scene_name, dataset_folder, display_images=False, display_time=500,
                               dataset_type=DatasetType.imperfect):
        """
        Load stereo pair images from scene folder

        Parameters:
            scene_name (string): Scene name to load data (use 'get_scene_list()'
                to see possible scenes) Must have already been downloaded to the
                dataset foldering using 'download_scene_data()'
            dataset_folder (string): Path to folder where dataset has been downloaded
            display_image (bool): Optional. Should the scene data be displayed when it is loaded.
            display_time (int): Optional. Length of time to display each image once loaded.
            image_suffix (string): Optional. Addition suffix to load alternate left views
                                                'E' means exposure changed between views
                                                'L' means lighting changed between views

        Returns:
            left_image (numpy): 2D image from left camera, loaded from scene data
            right_image (numpy): 2D image from right camera, loaded from scene data
        """
        left_image_filename = "im0.png"
        right_image_filename = "im1{}.png".format(Dataset.get_image_suffix(dataset_type))
        scene_year = Dataset.get_scene_year(scene_name)
        perfect_suffix = ""
        if scene_year == "2014":
            perfect_suffix = Dataset.get_perfect_suffix(dataset_type)
        # Define left and right image files in scene folder
        left_image_filepath = os.path.join(dataset_folder,
                                           scene_name+perfect_suffix,
                                           left_image_filename)
        right_image_filepath = os.path.join(dataset_folder,
                                            scene_name+perfect_suffix,
                                            right_image_filename)
        # Check left and right image files exist
        if not os.path.exists(left_image_filepath) or not os.path.exists(right_image_filepath):
            print("Left or right image file does not exist")
            print(left_image_filepath)
            print(right_image_filepath)
            raise PathNotFound(
                left_image_filepath+","+right_image_filepath,
                "Left or right image file does not exist")
        # Read left and right image files to numpy image
        left_image = cv2.imread(left_image_filepath, cv2.IMREAD_UNCHANGED)
        right_image = cv2.imread(right_image_filepath, cv2.IMREAD_UNCHANGED)
        if display_images:
            # Display left and right image files to OpenCV window
            left_image_resize = cv2.resize(left_image, dsize=(0, 0), fx=0.2, fy=0.2)
            right_image_resize = cv2.resize(right_image, dsize=(0, 0), fx=0.2, fy=0.2)
            cv2.imshow('image', left_image_resize)
            cv2.waitKey(display_time)
            cv2.imshow('image', right_image_resize)
            cv2.waitKey(display_time)
        return left_image, right_image

    @staticmethod
    def load_scene_data(scene_name, dataset_folder, display_images=False,
                        display_time=500, dataset_type=DatasetType.imperfect):
        """Load scene data from scene folder

        Parameters:
            scene_name (string): Scene name to load data (use 'get_scene_list()'
                to see possible scenes) Must have already been downloaded to the
                dataset foldering using 'download_scene_data()'
            dataset_folder (string): Path to folder where dataset has been downloaded
            display_image (bool): Optional. Should the scene data be displayed when it is loaded.
            display_time (int): Optional. Length of time to display each image once loaded.

        Returns:
            scene_data (TestData): Data loaded from scene folder
                (see TestData for details on this structure)
        """
        InvalidSceneName.validate_scene_list(scene_name, Dataset.get_scene_list())
        # Load stereo pair images from scene folder
        left_image, right_image = Dataset.load_scene_stereo_pair(
            scene_name, dataset_folder, display_images, display_time,
            dataset_type)

        # Load disparity image from scene folder
        disp_image = Dataset.load_scene_disparity(
            scene_name, dataset_folder, display_images, display_time,
            dataset_type)

        scene_year = Dataset.get_scene_year(scene_name)
        if scene_year == "2014":
            # Define path to calibration file in scene folder
            perfect_suffix = Dataset.get_perfect_suffix(dataset_type)
            cal_file = os.path.join(dataset_folder, scene_name+perfect_suffix, "calib.txt")
            # Get calibration data from calibration file
            cal_data = Dataset.load_cal(cal_file)
            ndisp = cal_data.ndisp
            # Calculate depth image from disparity using calibration file
            depth_image = Dataset.disp_to_depth(disp_image, cal_data.focal_length,
                                                cal_data.doffs, cal_data.baseline)
        else:
            depth_image = None
            # TODO: replace this with list of files that don't have cal data on website
            # however all images in training dataset without calibration files have a
            # ndisp = 256
            ndisp = 256

        return TestData(left_image, right_image, disp_image, depth_image, ndisp)

    @staticmethod
    def get_scene_list():
        """Return full list of scenes available"""
        return Dataset.STEREO_MIDDLEBURY_SCENES

    @staticmethod
    def get_training_scene_list():
        """Returns list of training scenes aviablable"""
        return Dataset.STEREO_MIDDLEBURY_TRAINING_SCENES

    @staticmethod
    def get_url_from_scene(scene_name, dataset_type=DatasetType.imperfect):
        """
        Get URL on middlebury servers for 2014 dataset for chosen scene

        Parameters:
            scene_name (string): Scene name to get url of

        Returns:
            url (string): Url for chosen scene name in middlebury stereo dataset (2014)
        """
        scene_year = Dataset.get_scene_year(scene_name)
        if scene_year == "2014":
            base_url = "http://vision.middlebury.edu/stereo/data/scenes2014/zip/"
            perfect_suffix = Dataset.get_perfect_suffix(dataset_type)
            url = base_url+scene_name+perfect_suffix+".zip"
        elif scene_year == "2003":
            scene_name = scene_name.lower()
            base_url = "https://vision.middlebury.edu/stereo/data/scenes2003/"
            url_suffix = "newdata/{}/{}-png-2.zip"
            url = base_url+url_suffix.format(scene_name, scene_name)
        elif scene_year == "2005":
            # Images are not stored in zip file for this year.
            # Will return base url for this scene instead
            url = "https://vision.middlebury.edu/stereo/data/scenes2005/FullSize/Art/"
        return url

    @staticmethod
    def bar_progress(current, total, *_):
        """
        Progress bar to display download progress in wget

        Parameters:
            current (int): current byte count
            total (int): total number of bytes
            *_: required by wget but is ignored in this function
        """
        base_progress_msg = "Downloading: %d%% [%d / %d] bytes"
        progress_message = base_progress_msg % (current / total * 100, current, total)
        # Don't use print() as it will print in new line every time.
        sys.stdout.write("\r" + progress_message)
        sys.stdout.flush()

    @staticmethod
    def get_scene_year(scene_name):
        """
        Get dataset year scene is from (2003, 2005, or 2014)

        Parameters:
            scene_name (string): scene name in dataset

        Return:
            scene_year (string): dataset year the scene is from (2003, 2005, or 2014)
        """
        if scene_name in Dataset.STEREO_MIDDLEBURY_SCENES_2014:
            return "2014"
        if scene_name in Dataset.STEREO_MIDDLEBURY_SCENES_2005:
            return "2005"
        if scene_name in Dataset.STEREO_MIDDLEBURY_SCENES_2003:
            return "2003"
        raise InvalidSceneName(scene_name)

    @staticmethod
    def download_scene_2005_data(scene_name, output_folder, url):
        """
        Download scene data for 2005 Middlebury data

        Parameters:
            scene_name (string): Scene to download from Middlesbury stereo dataset (2014)
            output_folder (string): Path to download scene data
            url (string): url to download from
        """
        scene_output_folder = os.path.join(output_folder, scene_name)
        # Check scene folder doesn't already exist
        if not os.path.exists(scene_output_folder):
            # Create scene folder
            os.makedirs(scene_output_folder)
            img0_filepath = os.path.join(scene_output_folder, "im0.png")
            img1_filepath = os.path.join(scene_output_folder, "im1L.png")
            disp_filepath = os.path.join(scene_output_folder, "disp0.png")
            # TODO: find cal data
            filepaths = [img0_filepath, img1_filepath, disp_filepath]
            base_url = url
            urls = ["Illum1/Exp1/view1.png", "Illum2/Exp1/view5.png", "disp1.png"]
            urls = [base_url + url for url in urls]
            print("Downloading from: "+base_url)
            for index, url in enumerate(urls):
                filepath = filepaths[index]
                # Check file doesn't already exist
                if not os.path.exists(filepath):
                    # download file from middlebury server
                    wget.download(url, filepath, bar=Dataset.bar_progress)
                else:
                    msg = "Image file for dataset already exists here,"
                    msg += "skipping download of file: " + filepath
                    print(msg)
        else:
            print("Dataset already exists here, skipping re-download of "+scene_name)

    @staticmethod
    def download_scene_2014_2003_dataset(scene_name, output_folder, url, scene_year, dataset_type):
        """
        Download scene data for 2014 or 2003 Middlebury data

        Parameters:
            scene_name (string): Scene to download from Middlesbury stereo dataset (2014)
            output_folder (string): Path to download scene data
            url (string): url to download from
            scene_year (string): '2014' or '2003'
            dataset_type (DatasetType): used to get scene suffix folder naming
        """
        # Get perfect suffix from dataset ('-imperfect' or '-perfect')
        scene_suffix = Dataset.get_perfect_suffix(dataset_type)
        # Download dataset from middlebury servers
        # Get name of scene data folder
        if scene_year == "2014":
            scene_output_folder = os.path.join(output_folder, scene_name+scene_suffix)
        elif scene_year == "2003":
            scene_output_folder = os.path.join(output_folder, scene_name)
        # Define destination name for zip file
        if scene_year == "2014":
            zip_filepath = os.path.join(output_folder, scene_name+scene_suffix+".zip")
        elif scene_year == "2003":
            zip_filepath = os.path.join(output_folder, scene_name+".zip")
        # Check scene folder doesn't already exist
        if not os.path.exists(scene_output_folder):
            # Check zip file doesn't already exist
            if not os.path.exists(zip_filepath):
                print("Downloading from: "+url)
                # download file from middlebury server
                wget.download(url, zip_filepath, bar=Dataset.bar_progress)
            else:
                msg = "Zip file for dataset already exists here,"
                msg += "skipping download of file: " + zip_filepath
                print(msg)
            print("Extracting zip...")
            # unzip downloaded file
            with zipfile.ZipFile(zip_filepath, "r") as zip_ref:
                zip_ref.extractall(output_folder)
            if scene_year == "2003":
                os.rename(
                    os.path.join(output_folder, scene_name.lower()), scene_output_folder)
                os.rename(
                    os.path.join(scene_output_folder, "im2.png"),
                    os.path.join(scene_output_folder, "im0.png"))
                os.rename(
                    os.path.join(scene_output_folder, "im6.png"),
                    os.path.join(scene_output_folder, "im1.png"))
                os.rename(
                    os.path.join(scene_output_folder, "disp2.png"),
                    os.path.join(scene_output_folder, "disp0.png"))
                os.remove(os.path.join(scene_output_folder, "disp6.png"))
            # removing zip file
            os.remove(zip_filepath)
        else:
            print("Dataset already exists here, skipping re-download of "+scene_name)

    @staticmethod
    def download_scene_data(scene_name, output_folder, dataset_type=DatasetType.imperfect):
        """
        Download scene data

        Parameters:
            scene_name (string): Scene to download from Middlesbury stereo dataset (2014)
            output_folder (string): Path to download scene data
        """
        # Check output folder exists
        if os.path.exists(output_folder):
            # clean-up tmp files from incomplete downloads
            tmp_files = glob.glob(os.path.join(output_folder, "*tmp"))
            for tmp_file in tmp_files:
                os.remove(tmp_file)
            # Get url from scene name
            url = Dataset.get_url_from_scene(scene_name, dataset_type)
            # get scene year (will change how data is downloaded)
            scene_year = Dataset.get_scene_year(scene_name)
            if scene_year == "2014" or scene_year == "2003":
                Dataset.download_scene_2014_2003_dataset(scene_name, output_folder,
                                                         url, scene_year, dataset_type)
            elif scene_year == "2005":
                Dataset.download_scene_2005_data(scene_name, output_folder, url)

        else:
            raise Exception('Output folder not found for storing datasets')



"""
Stereo Middlebury Evaluation package

This module is for loading the stereo Middlebury dataset loading
and includes tools evaluatating stereo matching algorithms
"""
import numpy as np
import cv2


def normalise_disp(disp_image):
    """
    Normalise disparity image

    Parameters:
        disp_image (numpy): disparity image
    Results:
        norm_disp_image (numpy): normalised disparity image
    """
    norm_disp_image = disp_image
    # remove negative disparities
    norm_disp_image[norm_disp_image < 0] = 0.0
    # Replace nan and inf values with zero
    norm_disp_image = np.nan_to_num(norm_disp_image, nan=0.0, posinf=0.0, neginf=0.0)
    # normalise image
    norm_disp_image = cv2.normalize(norm_disp_image, None, 0, 255, cv2.NORM_MINMAX)
    # convert to uint8
    norm_disp_image = norm_disp_image.astype(np.uint8)
    return norm_disp_image


def colormap_disp(disp_image):
    """
    Apply colormap to disparity

    Parameters:
        disp_image (numpy): disparity image
    Returns:
        colormap_disp_image (numpy): disparity image with colormap applied
    """
    colormap_disp_image = normalise_disp(disp_image)
    colormap_disp_image = cv2.applyColorMap(colormap_disp_image, cv2.COLORMAP_JET)
    return colormap_disp_image


# Helpful functions
def in_dictlist(key, value, my_dictlist):
    """Check if key value pair exists in dictionary"""
    for this in my_dictlist:
        if this[key] == value:
            return True
    return False


def puttext_multiline(img, text, org, font, font_scale, color, thickness,
                      line_type=cv2.LINE_AA, line_spacing=1.5, outline_color=None):
    """
    Draws multiline with an outline.
    """
    assert isinstance(text, str)

    uv_top_left = np.array(org, dtype=float)
    assert uv_top_left.shape == (2,)

    for line in text.splitlines():
        (_, height), _ = cv2.getTextSize(
            text=line,
            fontFace=font,
            fontScale=font_scale,
            thickness=thickness,
        )
        uv_bottom_left_i = uv_top_left + [0, height]
        org = tuple(uv_bottom_left_i.astype(int))
        if outline_color is not None:
            cv2.putText(
                img,
                text=line,
                org=org,
                fontFace=font,
                fontScale=font_scale,
                color=outline_color,
                thickness=thickness * 3,
                lineType=line_type,
            )
        cv2.putText(
            img,
            text=line,
            org=org,
            fontFace=font,
            fontScale=font_scale,
            color=color,
            thickness=thickness,
            lineType=line_type,
        )
        uv_top_left += [0, height * line_spacing]
    return img


def image_resize(image, width=None, height=None, inter=cv2.INTER_AREA):
    """
    Resize image based on height or width while maintaning aspect ratio
    Parameters:
        image (numpy): image to resize
        width (int): desired width of output image (can only use width or height, not both)
        height (int): desired height of output image (can only use width or height, not both)
        inter (int): OpenCV resize method (default: cv2.INTER_AREA)

    Returns:
        resized (numpy): resized image
    """
    if width is not None and height is not None:
        raise Exception("Cannot set both height and width")
    # initialize the dimensions of the image to be resized and
    # grab the image size
    dim = None
    (current_height, current_width) = image.shape[:2]

    # if both the width and height are None, then return the
    # original image
    if width is None and height is None:
        return image

    # check to see if the width is None
    if width is None:
        # calculate the ratio of the height and construct the
        # dimensions
        ratio = height / float(current_height)
        dim = (int(current_width * ratio), height)

    # otherwise, the height is None
    else:
        # calculate the ratio of the width and construct the
        # dimensions
        ratio = width / float(current_width)
        dim = (width, int(current_height * ratio))

    # resize the image
    resized = cv2.resize(image, dim, interpolation=inter)

    # return the resized image
    return resized



"""Exceptions"""
import os


class ImageSizeNotEqual(Exception):
    """Image size not equal exception"""
    def __str__(self):
        return "Image sizes must be equal"

    @staticmethod
    def validate(image_a, image_b):
        """
        Validate standard exception condition.
        Raises exception if validation fails.

        Paramaters:
            image_a (numpy): image to compare
            image_b (numpy): image to compare
        """
        if image_a.shape != image_b.shape:
            raise ImageSizeNotEqual()


class InvalidSceneName(Exception):
    """Invalid scene exception"""
    def __init__(self, scene_name):
        """
        Exception handelling for invalid scene name
        Parameters:
            scene_name (string): Scene name that failed
        """
        self.message = "Invalid scene name '{}'.".format(scene_name)
        super().__init__(self.message)

    def __str__(self):
        return self.message

    @staticmethod
    def validate_scene_list(scene_name, scene_list):
        """
        Validate standard exception condition.
        Raises exception if validation fails.

        Parameters:
            scene_name (string): scene name to test
            scene_list (list(string)): list of scene names to compare against
        """
        if scene_name not in scene_list:
            raise InvalidSceneName(scene_name)

    @staticmethod
    def validate_scene_info_list(scene_name, scene_info_list):
        """
        Validate standard exception condition.
        Raises exception if validation fails.

        Parameters:
            scene_name (string): scene name to test
            scene_info_list (list(SceneInfo)):
                list of scene info to compare against
        """
        for scene_info in scene_info_list:
            if scene_name == scene_info.scene_name:
                return
        raise InvalidSceneName(scene_name)


class MalformedPFM(Exception):
    """Malformed PFM file exception"""
    def __init__(self, message="Malformed PFM file"):
        """
        Exception handelling for malformed PFM file
        Parameters:
            message (string): Message to display in exception,
                will use default message if no message is provided
        """
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        """Overload of exception message"""
        return self.message


class PathNotFound(Exception):
    """Path not found exception"""
    def __init__(self, filepath, message="Path does not exist: "):
        """
        Exception handelling for path not found
        Parameters:
            filepath (string): Filepath that was not found.
                Will be displayed in exception message
            message (string): Message to display in exception,
                will use default message if no message is provided
        """
        self.filepath = filepath
        self.message = message
        super().__init__(self.message)

    @staticmethod
    def validate(filepath):
        """
        Validate standard exception condition.
        Raises exception if validation fails.

        Parameters:
            filepath (string): filepath to test
        """
        if not os.path.exists(filepath):
            raise PathNotFound(filepath)

    def __str__(self):
        """Overload of exception message"""
        return self.message+self.filepath



"""Structures"""


class DatasetType:
    """Dataset type"""
    imperfect = 'I'
    exposure_changed = 'E'
    lighting_changed = 'L'
    perfect = 'P'
    SCENE_CHANGED_TYPES = [exposure_changed, lighting_changed]


class CalibrationData:
    """
    Calibration data

    Used to make returning and accessing calibration data simple.
    """
    def __init__(self, width, height, c_x, c_y, focal_length, doffs, baseline, ndisp):
        """
        Initalisaiton of CalibrationData structure

        Parameters:
            c_x (float): Principle point in X
            c_y (float): Principle point in Y
            focal_length (float): focal length
            doffs (float): x-difference of principal points, doffs = cx1 - cx0
            baseline (float): baseline
        """
        self.width = width
        self.height = height
        self.c_x = c_x
        self.c_y = c_y
        self.focal_length = focal_length
        self.doffs = doffs
        self.baseline = baseline
        self.ndisp = ndisp


class TestData:
    """
    Test data

    Used to make returning and accessing scene data simple.
    """
    def __init__(self, left_image, right_image, disp_image, depth_image, ndisp):
        """
        Initalisaiton of TestData structure

        Parameters:
            left_image (numpy): 2D image from left camera in
                stereo pair
            right_image (numpy): 2D image from right camera in
                stereo pair
            disp_image (numpy): 2D disparity image (result of stereo matching)
            depth_image (numpy): 2D depth image (units: meters)
        """
        self.left_image = left_image
        self.right_image = right_image
        self.disp_image = disp_image
        self.depth_image = depth_image
        self.ndisp = ndisp


class SceneInfo:
    """
    Scene info
    """
    def __init__(self, scene_name, dataset_type, weight=1.0):
        """
        Initalisaiton of SceneInfo structure

        Parameters:
        """
        self.scene_name = scene_name
        self.dataset_type = dataset_type
        self.weight = weight

    def get_unique_name(self):
        scene_uniq_name = self.scene_name
        if self.dataset_type != DatasetType.imperfect:
            scene_uniq_name = self.scene_name+self.dataset_type
        return scene_uniq_name


class MatchData:
    """Match data"""

    class MatchResult:
        """Match result"""
        def __init__(self, left_image, right_image, ground_truth, match_result, match_time, ndisp=None):
            self.left_image = left_image
            self.right_image = right_image
            self.ground_truth = ground_truth
            self.match_result = match_result
            self.match_time = match_time
            self.ndisp = ndisp

    def __init__(self, scene_info, match_result):
        self.scene_info = scene_info
        self.match_result = match_result


class EvaluationData:
    """Evaluation data"""
    class EvaluationResult:
        """Evaluation result"""
        def __init__(self, metric, result, rank=None, ndisp=None):
            self.metric = metric
            self.result = result
            self.rank = rank
            self.ndisp = ndisp

    def __init__(self, scene_info, eval_result_list):
        self.scene_info = scene_info
        self.eval_result_list = eval_result_list



"""Setup module to build stereomideval module"""
from os.path import abspath, dirname, join, normpath, relpath
from shutil import rmtree
import glob
import setuptools
from setuptools import Command

with open("../README.md", "r") as fh:
    long_description = fh.read()

with open("../version.txt", "r") as fh:
    version = fh.read()


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    CLEAN_FILES = './build ./dist ./*.pyc ./*.tgz ./*.egg-info ./__pycache__'.split(' ')

    # Support the "all" option. Setuptools expects it in some situations.
    user_options = [
        ('all', 'a',
         "provided for compatibility, has no extra functionality")
    ]

    boolean_options = ['all']

    def __init__(self, dist):
        self.all = None
        super().__init__(dist)

    def initialize_options(self):
        """Inital options for clean command"""
        self.all = None

    def finalize_options(self):
        """Finalise options for clean command"""

    def run(self):
        """Run clean command"""
        script_path = normpath(abspath(dirname(__file__)))
        for path_spec in self.CLEAN_FILES:
            # Make paths absolute and relative to this path
            abs_paths = glob.glob(normpath(join(script_path, path_spec)))
            for path in [str(p) for p in abs_paths]:
                if not path.startswith(script_path):
                    # Die if path in CLEAN_FILES is absolute + outside this directory
                    raise ValueError("%s is not a path inside %s" % (path, script_path))
                print('removing %s' % relpath(path))
                rmtree(path)


setuptools.setup(
    name="stereo-mideval",
    version=version,
    author="Ben Knight",
    author_email="bknight@i3drobotics.com",
    description="Evaluation dataset and tools from Middlebury Stereo Evaulation data 2014.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/i3drobotics/stereomideval",
    packages=setuptools.find_packages(),
    package_dir={'stereomideval': 'stereomideval'},
    install_requires=[
        'numpy; python_version == "3.5"', 'numpy==1.19.3; python_version > "3.5"',
        'opencv-python', 'requests', 'wget', 'beautifulsoup4'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
    cmdclass={
        'clean': CleanCommand,
    }
)
